from django.core.validators import RegexValidator


hostname_validator = RegexValidator(
    r"^(?:[a-zA-Z0-9][a-zA-Z0-9\-]{0,63}\.)*$",
    "Only alphanumeric chars, hyphens and dots allowed",
)

zone_validator = RegexValidator(
    r"\.$",
    "Zone name must end with a dot",
)


import powerdns

from .utils import can_manage_record, get_managed_comment


class DnsRecord:
    def __init__(self, name:str, data:str, dns_type:str, zone_name:str, ttl:int):
        self.name = name.replace(zone_name, "")
        self.name = self.name.rstrip(".")
        self.data = data
        self.dns_type = dns_type
        self.ttl = ttl
        self.zone_name = zone_name

    @classmethod
    def from_pdns_record(cls, record:dict, zone:powerdns.interface.PDNSZone) -> tuple['DnsRecord']:
        dns_records = set()
        if not can_manage_record(record):
            return set()
        for content in record["records"]:
            dns_record = cls(
                name=record["name"],
                dns_type=record["type"],
                ttl=record["ttl"],
                data=content["content"],
                zone_name=zone.name,
            )
            dns_records.add(dns_record)
        return dns_records

    def as_rrset(self) -> powerdns.RRSet:
        return powerdns.RRSet(self.name, self.dns_type, [self.data], ttl=self.ttl, comments=get_managed_comment())

    def __hash__(self) -> int:
        return hash(tuple([self.name, self.dns_type, self.ttl, self.data, self.zone_name]))

    def __eq__(self, other: "DnsRecord") -> bool:
        return self.name == other.name and self.data == other.data and \
            self.dns_type == other.dns_type and self.ttl == other.ttl and \
            self.zone_name == other.zone_name

    def __repr__(self) -> str:
        return f"<DNSRecord: {self}>"

    def __str__(self) -> str:
        return f"{self.name}.{self.zone_name} {self.dns_type} {self.ttl} {self.data}"



from utilities.choices import ChoiceSet


class NamingIpChoices(ChoiceSet):
    key = "Zone.naming_ip_method"
    CHOICES = [
        ("netbox_powerdns_sync.naming.NamingIpDnsName", "Use IP DNS name field"),
        ("netbox_powerdns_sync.naming.NamingIpReverse", "Use PTR format for IP (without arpa domain)"),
    ]


class NamingDeviceChoices(ChoiceSet):
    key = "Zone.naming_device_method"
    CHOICES = [
        ("netbox_powerdns_sync.naming.NamingDeviceName", "Use device name only"),
        ("netbox_powerdns_sync.naming.NamingDeviceByInterface", "Use interface.device format"),
        ("netbox_powerdns_sync.naming.NamingDeviceByInterfacePrimary", "Use interface.device format and device only for primary IP"),
    ]


class NamingFgrpGroupChoices(ChoiceSet):
    key = "Zone.naming_fhrpgroup_method"
    CHOICES = [
        ("netbox_powerdns_sync.naming.NamingFGRPGroupName", "Use FHRP Group name only"),
    ]



from extras.plugins import PluginMenuItem, PluginMenuButton
from utilities.choices import ButtonColorChoices


apiserver_buttons = [
    PluginMenuButton(
        link='plugins:netbox_powerdns_sync:apiserver_add',
        title='Add',
        icon_class='mdi mdi-plus-thick',
        color=ButtonColorChoices.GREEN
    ),
]

zone_buttons = [
    PluginMenuButton(
        link='plugins:netbox_powerdns_sync:zone_add',
        title='Add',
        icon_class='mdi mdi-plus-thick',
        color=ButtonColorChoices.GREEN
    ),
]

menu_items = (
    PluginMenuItem(
        link='plugins:netbox_powerdns_sync:apiserver_list',
        link_text='API Servers',
        permissions=["netbox_powerdns_sync.view_apiserver"],
        buttons=apiserver_buttons,
    ),
    PluginMenuItem(
        link='plugins:netbox_powerdns_sync:zone_list',
        link_text='DNS Zones',
        permissions=["netbox_powerdns_sync.view_zone"],
        buttons=zone_buttons,
    ),
    PluginMenuItem(
        link='plugins:netbox_powerdns_sync:sync_jobs',
        link_text='Sync Jobs',
        permissions=["core.view_job"],
    ),
)



from extras.plugins import PluginConfig
from .version import __version__


class NetBoxPowerdnsSyncConfig(PluginConfig):
    name = "netbox_powerdns_sync"
    verbose_name = "NetBox PowerDNS sync"
    version = __version__
    description = "Sync DNS records in PowerDNS with NetBox"
    author = "Matej Vadnjal"
    author_email = "matej.vadnjal@arnes.si"
    base_url = "powerdns-sync"
    min_version = "3.6.0"
    default_settings = {
        "ttl_custom_field": None,
        "powerdns_managed_record_comment": "netbox-powerdns-sync",
        "post_save_enabled": False,
    }

    def ready(self):
        super().ready()
        import netbox_powerdns_sync.signals

config = NetBoxPowerdnsSyncConfig



import logging
import powerdns
import traceback
from datetime import timedelta
from django.db.models import Q

from core.choices import JobStatusChoices
from core.models import Job
from dcim.models import Device, Interface
from extras.choices import LogLevelChoices
from ipam.models import IPAddress, FHRPGroup
from netbox_powerdns_sync.constants import FAMILY_TYPES, PTR_TYPE
from virtualization.models import VirtualMachine, VMInterface

from .exceptions import *
from .models import ApiServer, Zone
from .naming import generate_fqdn
from .record import DnsRecord
from .utils import get_ip_ttl, make_dns_label, make_canonical


logger = logging.getLogger("netbox.netbox_powerdns_sync.jobs")


class JobLoggingMixin:
    def log(self, level: str, msg: str) -> None:
        data = self.job.data or {}
        logs = data.get("log", [])
        logs.append({
            "message": msg,
            "status": level,
        })
        data["log"] = logs
        self.job.data = data
    
    def log_debug(self, msg: str) -> None:
        logger.debug(msg)
        self.log(LogLevelChoices.LOG_DEFAULT, msg)

    def log_success(self, msg: str) -> None:
        logger.info(msg)
        self.log(LogLevelChoices.LOG_SUCCESS, msg)

    def log_info(self, msg: str) -> None:
        logger.info(msg)
        self.log(LogLevelChoices.LOG_INFO, msg)

    def log_warning(self, msg: str) -> None:
        logger.warning(msg)
        self.log(LogLevelChoices.LOG_WARNING, msg)

    def log_failure(self, msg: str) -> None:
        logger.error(msg)
        self.log(LogLevelChoices.LOG_FAILURE, msg)


class PowerdnsTask(JobLoggingMixin):
    def __init__(self, job: Job) -> None:
        self.job = job
        self.init_attrs()
    
    def init_attrs(self):
        self.fqdn : str = ""
        self.forward_zone : Zone = None
        self.reverse_zone : Zone = None
        self.make_fqdn_ran : bool = False

    def get_pdns_servers_for_zone(self, zone_name:str) -> list[ApiServer]:
        if not zone_name:
            return []
        zone = Zone.objects.get(name=zone_name)
        return zone.api_servers.enabled().all()

    def add_to_output(self, row):
        if not self.job.data:
            self.job.data = dict()
        if "output" not in self.job.data:
            self.job.data["output"] = []
        self.job.data["output"].append(row)

    def make_name_from_interface(self, interface: Interface|VMInterface, host: Device|VirtualMachine) -> str:
        name = host.name
        name = ".".join(map(make_dns_label, name.split(".")))
        if self.ip != host.primary_ip4 and self.ip != host.primary_ip6:
            name = make_dns_label(interface.name) + "." + name
        return name

    def make_fqdn(self) -> str:
        """ Determines FQDN and sets forward zone """
        if self.make_fqdn_ran:
            return self.fqdn
        self.make_fqdn_ran = True
        if self.determine_forward_zone():
            self.fqdn = generate_fqdn(self.ip, self.forward_zone)
        return self.fqdn

    def determine_forward_zone(self):
        # determine zone from any FQDN names
        name = None
        if self.ip.dns_name:
            name = self.ip.dns_name
        elif isinstance(self.ip.assigned_object, Interface):
            name = self.ip.assigned_object.device.name
        elif isinstance(self.ip.assigned_object, VMInterface):
            name = self.ip.assigned_object.virtual_machine.name
        elif isinstance(self.ip.assigned_object, FHRPGroup):
            name = self.ip.assigned_object.name
        if name:
            self.forward_zone = Zone.get_best_zone(name)
        # determine zone by matching tags or roles
        if not self.forward_zone:
            self.forward_zone = Zone.match_ip(self.ip).first()
        return self.forward_zone

    def make_reverse_domain(self) -> str|None:
        """ Returns reverse domain name """
        return make_canonical(self.ip.address.ip.reverse_dns)

    def create_record(self, dns_record: DnsRecord) -> None:
        servers = self.get_pdns_servers_for_zone(dns_record.zone_name)
        if not servers:
            raise PowerdnsSyncNoServers(f"No valid servers found for zone {dns_record.zone_name}")
        for api_server in self.get_pdns_servers_for_zone(dns_record.zone_name):
            zone = api_server.api.get_zone(dns_record.zone_name)
            if not zone:
                raise PowerdnsSyncServerZoneMissing(
                    f"Zone {dns_record.zone_name} not found on server {api_server}"
                )
            self.add_to_output({"action": "CREATE", "rr": str(dns_record), "zone": str(zone), "server": str(api_server)})
            zone.create_records([dns_record.as_rrset()])

    def delete_record(self, dns_record: DnsRecord) -> None:
        servers = self.get_pdns_servers_for_zone(dns_record.zone_name)
        if not servers:
            raise PowerdnsSyncNoServers(f"No valid servers found for zone {dns_record.zone_name}")
        for api_server in self.get_pdns_servers_for_zone(dns_record.zone_name):
            zone = api_server.api.get_zone(dns_record.zone_name)
            if not zone:
                raise PowerdnsSyncServerZoneMissing(
                    f"Zone {dns_record.zone_name} not found on server {api_server}"
                )
            self.add_to_output({"action": "DELETE", "rr": str(dns_record), "zone": str(zone), "server": str(api_server)})
            zone.delete_records([dns_record.as_rrset()])


class PowerdnsTaskIP(PowerdnsTask):
    def __init__(self, job: Job) -> None:
        super().__init__(job)
        self.ip : IPAddress = job.object

    @classmethod
    def run_update_ip(cls, job: Job, *args, **kwargs) -> None:
        task = cls(job)
        if job.object_id and not job.object:
            task.job.start()
            task.log_warning("No IP Address object given. IP was probably removed or DB transaction aborted, nothing to do.")
            task.job.terminate(status=JobStatusChoices.STATUS_COMPLETED)
            return
        try:
            task.log_debug("Starting task")
            task.job.start()
            task.log_debug("Creating forward record")
            task.create_forward()
            task.log_debug("Creating reverse record")
            task.create_reverse()
            task.log_success("Finished")
            task.job.terminate()
        except Exception as e:
            task.log_failure(f"error {e}")
            task.job.data = task.job.data or dict()
            task.job.data["exception"] = str(e)
            task.job.terminate(status=JobStatusChoices.STATUS_ERRORED)
            raise e

    def create_forward(self) -> None:
        self.make_fqdn()
        if not self.forward_zone:
            raise PowerdnsSyncNoZoneFound(f"No forward zone found for IP:{self.ip}")
        if not self.fqdn:
            raise PowerdnsSyncNoNameFound(f"No forward name for IP:{self.ip} (zone:{self.forward_zone})")
        name = self.fqdn.replace(self.forward_zone.name, "").rstrip(".")
        dns_record = DnsRecord(
            name=name,
            dns_type=FAMILY_TYPES[self.ip.family],
            data=str(self.ip.address.ip),
            ttl=get_ip_ttl(self.ip) or self.forward_zone.default_ttl,
            zone_name=self.forward_zone.name,
        )
        self.log_info(f"Forward record: {dns_record}")
        self.create_record(dns_record)
        self.log_info(f"Forward record created")

    def create_reverse(self) -> None:
        self.make_fqdn()
        if not self.fqdn:
            raise PowerdnsSyncNoNameFound(f"No forward name for IP:{self.ip}")
        reverse_fqdn = self.make_reverse_domain()
        self.reverse_zone = Zone.get_best_zone(reverse_fqdn)
        if not self.reverse_zone:
            self.log_warning(f"No reverse zone for IP:{self.ip} fqdn:{self.fqdn} Skipping")
            return
        name = reverse_fqdn.replace(self.reverse_zone.name, "").rstrip(".")
        dns_record = DnsRecord(
            name=name,
            dns_type=PTR_TYPE,
            data=self.fqdn,
            ttl=get_ip_ttl(self.ip) or self.reverse_zone.default_ttl,
            zone_name=self.reverse_zone.name,
        )
        self.log_info(f"Reverse record {dns_record}")
        self.create_record(dns_record)
        self.log_info(f"Reverse record created")


class PowerdnsTaskFullSync(PowerdnsTask):
    def __init__(self, job: Job) -> None:
        super().__init__(job)
        self.zone : Zone = job.object

    @classmethod
    def run_full_sync(cls, job: Job, *args, **kwargs) -> None:
        task = cls(job)

        try:
            task.log_debug(f"Starting sync for zone {task.zone}")
            task.job.start()
            if not task.zone.enabled:
                task.log_warning(f"Zone {task.zone} is disabled for updates, not syncing")
                task.job.terminate()
                return
            netbox_records = task.load_netbox_records()
            pdns_records = task.load_pdns_records()
            task.log_info(f"Found record count: netbox:{len(netbox_records)} pdns:{len(pdns_records)}")
            to_delete = pdns_records - netbox_records
            to_create = netbox_records - pdns_records
            task.log_info(f"Record change count: to_delete:{len(to_delete)} to_create:{len(to_create)}")
            for record in to_delete:
                task.delete_record(record)
            for record in to_create:
                task.create_record(record)
            task.log_success("Finished")
            task.job.terminate()
        except PowerdnsSyncNoServers as e:
            task.log_failure(str(e))
            task.job.data = task.job.data or dict()
            task.job.terminate(status=JobStatusChoices.STATUS_ERRORED)
        except Exception as e:
            stacktrace = traceback.format_exc()
            task.log_failure(f"An exception occurred: `{type(e).__name__}: {e}`\n```\n{stacktrace}\n```")
            task.job.data = task.job.data or dict()
            task.job.terminate(status=JobStatusChoices.STATUS_ERRORED)

        # Schedule the next job if an interval has been set
        if job.interval:
            new_scheduled_time = job.scheduled + timedelta(minutes=job.interval)
            Job.enqueue(
                cls.run_full_sync,
                instance=job.object,
                name=job.name,
                user=job.user,
                schedule_at=new_scheduled_time,
                interval=job.interval,
            )

    def get_addresses(self):
        """ Get IPAddress objects that could have DNS records """
        zone_canonical = self.zone.name
        zone_domain = self.zone.name.rstrip(".")
        # filter for FQDN names (ip.dns_name, Device, VM, FHRPGroup)
        query_zone = Q(dns_name__endswith=zone_canonical)|Q(dns_name__endswith=zone_domain)
        query_zone |= Q(interface__device__name__endswith=zone_canonical)|Q(interface__device__name__endswith=zone_domain)
        query_zone |= Q(vminterface__virtual_machine__name__endswith=zone_canonical)|Q(vminterface__virtual_machine__name__endswith=zone_domain)
        query_zone |= Q(fhrpgroup__name__endswith=zone_canonical)|Q(fhrpgroup__name__endswith=zone_domain)
        # filter for matchers (tags & roles)
        query_zone |= Q(tags__in=self.zone.match_ipaddress_tags.all())
        query_zone |= Q(interface__tags__in=self.zone.match_interface_tags.all())
        query_zone |= Q(vminterface__tags__in=self.zone.match_interface_tags.all())
        query_zone |= Q(interface__device__tags__in=self.zone.match_device_tags.all())
        query_zone |= Q(vminterface__virtual_machine__tags__in=self.zone.match_device_tags.all())
        query_zone |= Q(fhrpgroup__tags__in=self.zone.match_fhrpgroup_tags.all())
        query_zone |= Q(interface__device__role__in=self.zone.match_device_roles.all())
        query_zone |= Q(vminterface__virtual_machine__role__in=self.zone.match_device_roles.all())
        results = IPAddress.objects.filter(query_zone)
        if self.zone.match_interface_mgmt_only:
            results = results.filter(interface__mgmt_only=True)
        return results

    def load_netbox_records(self) -> set[DnsRecord]:
        records = set()
        ip: IPAddress
        ip_addresses = self.get_addresses()
        self.log_info(f"Found {ip_addresses.count()} matching addresses to check")
        for ip in ip_addresses:
            self.init_attrs()
            self.ip = ip
            self.make_fqdn()
            if not self.forward_zone:
                self.log_info(f"No matching forward zone found for IP:{ip}. Skipping")
                continue
            if not self.fqdn:
                self.log_info(f"No FQDN could be determined for IP:{ip} (zone:{self.forward_zone}). Skipping")
                continue
            if self.forward_zone == self.zone:
                name = self.fqdn.replace(self.forward_zone.name, "").rstrip(".")
                records.add(DnsRecord(
                    name=name,
                    data=str(ip.address.ip),
                    dns_type=FAMILY_TYPES.get(ip.family),
                    zone_name=self.forward_zone.name,
                    ttl=get_ip_ttl(ip) or self.forward_zone.default_ttl,
                ))
            if self.zone.is_reverse:
                reverse_fqdn = self.make_reverse_domain()
                self.reverse_zone = Zone.get_best_zone(reverse_fqdn)
                if not self.reverse_zone:
                    self.log_info(f"No matching reverse zone for {ip} ({self.fqdn}). Skipping")
                    continue
                if self.reverse_zone == self.zone:
                    name = reverse_fqdn.replace(self.reverse_zone.name, "").rstrip(".")
                    records.add(DnsRecord(
                        name=name,
                        data=self.fqdn,
                        dns_type=PTR_TYPE,
                        zone_name=self.reverse_zone.name,
                        ttl=get_ip_ttl(ip) or self.reverse_zone.default_ttl,
                    ))
        return records

    def load_pdns_records(self) -> set[DnsRecord]:
        flat_records = set()
        checked_types = [PTR_TYPE] + list(FAMILY_TYPES.values())
        servers = self.get_pdns_servers_for_zone(self.zone.name)
        if not servers:
            raise PowerdnsSyncNoServers(f"No valid servers found for zone {self.zone}")
        for api_server in servers:
            pdns_zone = api_server.api.get_zone(self.zone.name)
            if not pdns_zone:
                raise PowerdnsSyncServerZoneMissing(
                    f"Zone {self.zone.name} not found on server {api_server}"
                )
            for record in pdns_zone.records:
                if record["type"] not in checked_types:
                    continue
                flat_records.update(DnsRecord.from_pdns_record(record, pdns_zone))
        return flat_records



import django_tables2 as tables
from core.tables import JobTable
from netbox.tables import NetBoxTable, columns

from . import models


__all__ = (
    "ApiServerTable",
    "SyncJobTable",
    "ZoneTable",
)

SYNC_URL_ID = """<a href="{% url 'plugins:netbox_powerdns_sync:sync_result' job_pk=value %}">{{ value }}</a>"""
SYNC_URL_NAME = """<a href="{% url 'plugins:netbox_powerdns_sync:sync_result' job_pk=record.id %}">{{ value }}</a>"""
DEVICE_ROLE_COLUMN = """
{% for role in value.all %}
    <span class="badge" style="color: {{ role.color|fgcolor }}; background-color: #{{ role.color }}">
        <a href="{{ role.get_absolute_url }}">{{ role }}</a>
    </span>
{% endfor %}
"""
ZONE_EXTRA_BUTTONS = """
{% if record.enabled %}
    <a href="{% url 'plugins:netbox_powerdns_sync:sync_schedule' %}?zones={{ record.pk }}" class="btn btn-sm btn-primary" title="Schedule full sync for zone">
        <span class="mdi mdi-sync" aria-hidden="true"></span>
    </a>
{% else %}
    <a href="#" class="btn btn-sm btn-primary disabled" title="Schedule full sync for zone">
        <span class="mdi mdi-sync" aria-hidden="true"></span>
    </a>
{% endif %}
"""


class ApiServerTable(NetBoxTable):
    name = tables.Column(
        linkify=True
    )
    zone_count = columns.LinkedCountColumn(
        viewname="plugins:netbox_powerdns_sync:zone_list",
        url_params={"api_server_id": "pk"},
        verbose_name="Zones"
    )
    enabled = columns.BooleanColumn()
    tags = columns.TagColumn(
        url_name="plugins:netbox_powerdns_sync:apiserver_list"
    )

    class Meta(NetBoxTable.Meta):
        model = models.ApiServer
        fields = (
            "pk", "id", "name", "zone_count", "description", "enabled", "api_url",
            "tags", "actions", "created", "last_updated",
        )
        default_columns = ("pk", "name", "zone_count", "description", "enabled")


class SyncJobTable(JobTable):
    id = tables.TemplateColumn(
        template_code=SYNC_URL_ID,
    )
    name = tables.TemplateColumn(
        template_code=SYNC_URL_NAME,
    )

    class Meta(JobTable.Meta):
        pass


class ZoneTable(NetBoxTable):
    name = tables.Column(
        linkify=True
    )
    api_server_count = columns.LinkedCountColumn(
        viewname="plugins:netbox_powerdns_sync:apiserver_list",
        url_params={"zone_id": "pk"},
        verbose_name="API Servers"
    )
    enabled = columns.BooleanColumn()
    is_default = columns.BooleanColumn()
    is_reverse = columns.BooleanColumn(
        verbose_name="Reverse zone",
        orderable=False,
    )
    match_ipaddress_tags = columns.TagColumn(
        url_name="ipam:ipaddress_list"
    )
    match_interface_tags = columns.TagColumn(
        url_name="dcim:interface_list"
    )
    match_device_tags = columns.TagColumn(
        url_name="dcim:device_list"
    )
    match_fhrpgroup_tags = columns.TagColumn(
        url_name="ipam:fhrpgroup_list"
    )
    match_device_roles = columns.TemplateColumn(
        template_code=DEVICE_ROLE_COLUMN
    )
    match_interface_mgmt_only = columns.BooleanColumn()
    tags = columns.TagColumn(
        url_name="plugins:netbox_powerdns_sync:zone_list",
    )
    actions = columns.ActionsColumn(
        extra_buttons=ZONE_EXTRA_BUTTONS,
    )

    class Meta(NetBoxTable.Meta):
        model = models.Zone
        fields = (
            "pk", "id", "name", "api_server_count", "description",  "enabled",
            "is_default", "is_reverse", "default_ttl", "match_ipaddress_tags",
            "match_interface_tags", "match_device_tags", "match_fhrpgroup_tags",
            "match_device_roles", "match_interface_mgmt_only", "naming_ip_method",
            "naming_device_method", "naming_fgrpgroup_method", "tags", "actions",
            "created", "last_updated",
        )
        default_columns = ("pk", "name", "enabled", "is_default", "default_ttl")



from django import forms
from django.core.exceptions import ValidationError
from netbox.forms import NetBoxModelForm
from utilities.forms import add_blank_choice

from ..choices import NamingDeviceChoices, NamingFgrpGroupChoices, NamingIpChoices
from ..models import ApiServer, Zone
from ..utils import is_reverse

__all__ = (
    "ApiServerForm",
    "ZoneForm",
)


class ApiServerForm(NetBoxModelForm):
    fieldsets = (
        ("API Server", (
            "name", "api_url", "api_token", "description", "enabled", "tags",
        )),
    )

    class Meta:
        model = ApiServer
        fields = [
            "name", "api_url", "api_token", "description", "enabled", "tags",
        ]


class ZoneForm(NetBoxModelForm):
    naming_ip_method = forms.ChoiceField(
        choices=add_blank_choice(NamingIpChoices),
        required=False,
        label='From IP',
        help_text="How to construct DNS name from IP address. Leave blank to ignore.",
    )
    naming_device_method = forms.ChoiceField(
        choices=add_blank_choice(NamingDeviceChoices),
        required=False,
        label='From device',
        help_text="How to construct DNS name from device IP is assigned to. Leave blank to ignore.",
    )
    naming_fgrpgroup_method = forms.ChoiceField(
        choices=add_blank_choice(NamingFgrpGroupChoices),
        required=False,
        label='From FHRP Group',
        help_text="How to construct DNS name from FHRP Group IP is assigned to. Leave blank to ignore.",
    )

    fieldsets = (
        ("DNS Zone", (
            "name", "description", "enabled", "api_servers", "is_default",
            "default_ttl",
        )),
        ("Matchers", (
            "match_ipaddress_tags", "match_interface_tags",
            "match_device_tags", "match_fhrpgroup_tags", "match_device_roles",
            "match_interface_mgmt_only", 
        )),
        ("Naming methods", (
            "naming_ip_method", "naming_device_method", "naming_fgrpgroup_method",
        )),
        ("General", ("tags",)),
    )

    class Meta:
        model = Zone
        fields = [
            "name", "description", "enabled", "api_servers", "is_default",
            "default_ttl", "match_ipaddress_tags", "match_interface_tags",
            "match_device_tags", "match_fhrpgroup_tags", "match_device_roles",
            "match_interface_mgmt_only", "naming_ip_method", "naming_device_method",
            "naming_fgrpgroup_method", "tags",
        ]

    def clean(self):
        super().clean()
        self.clean_match_tags(self.cleaned_data)
        self.clean_match_roles(self.cleaned_data)
        self.clean_naming_methods(self.cleaned_data)

    def clean_match_tags(self, data):
        if not is_reverse(data["name"]):
            return
        fields = (
            "match_ipaddress_tags",
            "match_interface_tags",
            "match_device_tags",
            "match_fhrpgroup_tags",
        )
        for f in fields:
            if data.get(f):
                self.add_error(f, "Cannot set match tags for reverse zone")

    def clean_match_roles(self, data):
        if not is_reverse(data["name"]):
            return
        if data.get("match_device_roles"):
            self.add_error("match_device_roles", "Cannot set match roles for reverse zone")

    def clean_naming_methods(self, data):
        if not is_reverse(data["name"]):
            return
        fields = (
            "naming_ip_method",
            "naming_device_method",
            "naming_fgrpgroup_method",
        )
        for f in fields:
            if data.get(f):
                self.add_error(f, "Cannot set naming methods for reverse zone")



# Generated by Django 4.1.9 on 2023-07-24 10:08

import django.core.validators
from django.db import migrations, models
import taggit.managers
import utilities.json


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('dcim', '0172_larger_power_draw_values'),
        ('extras', '0092_delete_jobresult'),
    ]

    operations = [
        migrations.CreateModel(
            name='ApiServer',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('name', models.CharField(max_length=100, unique=True)),
                ('description', models.CharField(blank=True, max_length=200)),
                ('enabled', models.BooleanField(default=True)),
                ('api_url', models.URLField(unique=True)),
                ('api_token', models.CharField(max_length=200)),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'verbose_name': 'API Server',
                'ordering': ('name',),
            },
        ),
        migrations.CreateModel(
            name='Zone',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False)),
                ('created', models.DateTimeField(auto_now_add=True, null=True)),
                ('last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('custom_field_data', models.JSONField(blank=True, default=dict, encoder=utilities.json.CustomFieldJSONEncoder)),
                ('name', models.CharField(max_length=200, unique=True, validators=[django.core.validators.MinLengthValidator(3), django.core.validators.RegexValidator('^(?:[a-zA-Z0-9][a-zA-Z0-9\\-]{0,63}\\.)*$', 'Only alphanumeric chars, hyphens and dots allowed'), django.core.validators.RegexValidator('\\.$', 'Zone name must end with a dot')])),
                ('description', models.CharField(blank=True, max_length=200)),
                ('enabled', models.BooleanField(default=True)),
                ('is_default', models.BooleanField(default=False)),
                ('default_ttl', models.IntegerField(default=3600)),
                ('match_interface_mgmt_only', models.BooleanField(default=False)),
                ('naming_ip_method', models.CharField(default=None, max_length=200, null=True)),
                ('naming_device_method', models.CharField(default=None, max_length=200, null=True)),
                ('naming_fgrpgroup_method', models.CharField(default=None, max_length=200, null=True)),
                ('api_servers', models.ManyToManyField(related_name='zones', to='netbox_powerdns_sync.apiserver')),
                ('match_device_roles', models.ManyToManyField(blank=True, related_name='+', to='dcim.devicerole')),
                ('match_device_tags', models.ManyToManyField(blank=True, related_name='+', to='extras.tag')),
                ('match_fhrpgroup_tags', models.ManyToManyField(blank=True, related_name='+', to='extras.tag')),
                ('match_interface_tags', models.ManyToManyField(blank=True, related_name='+', to='extras.tag')),
                ('match_ipaddress_tags', models.ManyToManyField(blank=True, related_name='+', to='extras.tag')),
                ('tags', taggit.managers.TaggableManager(through='extras.TaggedItem', to='extras.Tag')),
            ],
            options={
                'ordering': ('name',),
            },
        ),
        migrations.AddConstraint(
            model_name='zone',
            constraint=models.UniqueConstraint(condition=models.Q(('is_default', True)), fields=('is_default',), name='unique_is_default', violation_error_message='Only one zone can be set as default'),
        ),
    ]



from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.http import HttpResponseForbidden
from django.shortcuts import get_object_or_404, redirect, render
from django.views.generic import View
from core.models import Job
from utilities.htmx import is_htmx
from utilities.rqworker import get_workers_for_queue
from utilities.utils import normalize_querydict
from utilities.views import ContentTypePermissionRequiredMixin

from ..constants import JOB_NAME_DEVICE, JOB_NAME_INTERFACE, JOB_NAME_IP, JOB_NAME_SYNC
from ..jobs import PowerdnsTaskFullSync
from ..forms import ZoneScheduleForm
from ..models import Zone
from ..tables import SyncJobTable

__all__ = (
    "SyncJobsView",
    "SyncResultView",
    "SyncScheduleView",
)


class aaaSyncRunView(ContentTypePermissionRequiredMixin, View):

    def get_required_permission(self):
        return "extras.view_script"

    def get(self, request, pk):
        if not request.user.has_perm("extras.run_script"):
            return HttpResponseForbidden()

        zone = get_object_or_404(Zone.objects.restrict(request.user), pk=pk)
        if not zone.enabled:
            messages.error(request, f"Unable to sync disabled zone {zone}")
        elif not get_workers_for_queue("default"):
            messages.error(request, "Unable to run script: RQ worker process not running.")
        else:
            job = Job.enqueue(
                PowerdnsTaskFullSync.run_full_sync,
                instance=zone,
                name=JOB_NAME_SYNC,
                user=request.user,
            )
            return redirect("plugins:netbox_powerdns_sync:sync_result", job_pk=job.pk)

        return redirect(zone)


class SyncJobsView(ContentTypePermissionRequiredMixin, View):

    def get_required_permission(self):
        return "extras.view_script"

    def get(self, request):
        query = Q(app_label="netbox_powerdns_sync", model="zone")|Q(app_label="ipam", model="ipaddress")
        object_types = ContentType.objects.filter(query)
        jobs = Job.objects.filter(
            object_type__in=object_types,
            name__in=(JOB_NAME_DEVICE, JOB_NAME_INTERFACE, JOB_NAME_IP, JOB_NAME_SYNC),
        )
        jobs_table = SyncJobTable(
            data=jobs,
            orderable=False,
            user=request.user
        )
        jobs_table.configure(request)

        return render(request, "netbox_powerdns_sync/syncs.html", {
            "table": jobs_table,
            "tab": "jobs",
        })


class SyncResultView(ContentTypePermissionRequiredMixin, View):

    def get_required_permission(self):
        return "extras.view_script"

    def get(self, request, job_pk):
        #object_type = ContentType.objects.get_by_natural_key(app_label="extras", model="scriptmodule")
        #job = get_object_or_404(Job.objects.all(), pk=job_pk, object_type=object_type)
        job = get_object_or_404(Job.objects.all(), pk=job_pk)

        #module = job.object
        #script = module.scripts[job.name]()

        # If this is an HTMX request, return only the result HTML
        if is_htmx(request):
            response = render(request, "netbox_powerdns_sync/htmx/sync_result.html", {
                #"script": script,
                "job": job,
            })
            if job.completed or not job.started:
                response.status_code = 286
            return response

        return render(request, "netbox_powerdns_sync/sync_result.html", {
            #"script": script,
            "job": job,
        })


class SyncScheduleView(View):
    def get(self, request):
        scheduled_jobs = Job.objects.filter(status="scheduled", name=JOB_NAME_SYNC)
        jobs_table = SyncJobTable(
            data=scheduled_jobs,
            orderable=False,
            user=request.user
        )
        jobs_table.columns.show("scheduled")
        jobs_table.columns.show("interval")
        jobs_table.columns.show("object")
        jobs_table.configure(request)

        form = ZoneScheduleForm(initial=normalize_querydict(request.GET))

        return render(request, "netbox_powerdns_sync/sync_schedule.html", {
            "jobs_table": jobs_table,
            "form": form,
        })

    def post(self, request):
        form = ZoneScheduleForm(request.POST, request.FILES)
        
        if not get_workers_for_queue("default"):
            messages.error(request, "Unable to run script: RQ worker process not running.")
        elif form.is_valid():
            for zone in form.cleaned_data["zones"]:
                Job.enqueue(
                    PowerdnsTaskFullSync.run_full_sync,
                    instance=zone,
                    name=JOB_NAME_SYNC,
                    user=request.user,
                    schedule_at=form.cleaned_data.get("_schedule_at"),
                    interval=form.cleaned_data.get("_interval"),
                )
                messages.success(request, f"Scheduled sync job for zone {zone}")

        return redirect("plugins:netbox_powerdns_sync:sync_jobs")



PLUGIN_NAME = "netbox_powerdns_sync"
FAMILY_TYPES = {
    4: "A",
    6: "AAAA",
}
PTR_TYPE = "PTR"
PTR_ZONE_SUFFIXES = (
    "in-addr.arpa.",
    "ip6.arpa."
)

JOB_NAME_IP = "PowerDNS IP Address update"
JOB_NAME_INTERFACE = "PowerDNS Interface update"
JOB_NAME_DEVICE = "PowerDNS Device update"
JOB_NAME_SYNC = "PowerDNS zone sync"



import powerdns
from django.contrib.contenttypes.models import ContentType
from django.core.validators import MinLengthValidator
from django.db import models
from django.forms import ValidationError
from django.urls import reverse
from taggit.managers import TaggableManager
from core.models import Job
from dcim.models import DeviceRole, Interface
from ipam.models import IPAddress, FHRPGroup
from netbox.models import NetBoxModel
from extras.models import Tag
from virtualization.models import VMInterface

from .choices import NamingFgrpGroupChoices, NamingDeviceChoices, NamingIpChoices
from .constants import JOB_NAME_SYNC
from .querysets import EnabledQuerySet, ZoneQuerySet
from .utils import get_ip_host, make_canonical, is_reverse
from .validators import hostname_validator, zone_validator


__all__ = (
    "ApiServer",
    "Zone",
)


class ApiServer(NetBoxModel):
    name = models.CharField(
        max_length=100,
        blank=False,
        null=False,
        unique=True,
    )
    description = models.CharField(
        max_length=200,
        blank=True
    )
    enabled = models.BooleanField(
        null=False,
        default=True,
    )
    api_url = models.URLField(
        verbose_name="API URL",
        help_text="Base URL to PowerDNS REST API",
        max_length=200,
        unique=True,
    )
    api_token = models.CharField(
        verbose_name="API Token",
        max_length=200,
    )

    objects = EnabledQuerySet.as_manager()

    clone_fields = (
        "name", "api_url", "api_token", "description", "enabled", "tags",
    )

    class Meta:
        ordering = ("name",)
        verbose_name = "API Server"

    def get_absolute_url(self):
        return reverse("plugins:netbox_powerdns_sync:apiserver", args=[self.pk])

    def __str__(self):
        return self.name

    @property
    def api(self) -> powerdns.PDNSEndpoint|None:
        if not self.api_url or not self.api_url:
            return None
        api_client = powerdns.PDNSApiClient(
            api_endpoint=self.api_url,
            api_key=self.api_token,
        )
        return powerdns.PDNSEndpoint(api_client).servers[0]


class Zone(NetBoxModel):
    name = models.CharField(
        help_text="Domain name of zone. Must be fully qualified.",
        max_length=200,
        blank=False,
        null=False,
        unique=True,
        validators=[MinLengthValidator(3), hostname_validator, zone_validator],
    )
    description = models.CharField(
        max_length=200,
        blank=True
    )
    enabled = models.BooleanField(
        null=False,
        default=True,
    )
    api_servers = models.ManyToManyField(
        to=ApiServer,
        verbose_name="API Servers",
        related_name="zones",
    )
    is_default = models.BooleanField(
        help_text="This zone is to be used as default if no other zone matches.",
        null=False,
        default=False,
    )
    default_ttl = models.IntegerField(
        help_text="TTL to use for new records if not set by TTL custom field",
        verbose_name="Default TTL",
        default=3600,
    )
    match_ipaddress_tags = models.ManyToManyField(
        to=Tag,
        help_text="Use this zone for IP addresses with these tags",
        verbose_name="Match IP address tags",
        related_name="+",
        blank=True,
    )
    match_interface_tags = models.ManyToManyField(
        to=Tag,
        help_text="Use this zone for interfaces or VM interfaces with these tags",
        verbose_name="Match interface tags",
        related_name="+",
        blank=True,
    )
    match_device_tags = models.ManyToManyField(
        to=Tag,
        help_text="Use this zone for devices or VMs with these tags",
        verbose_name="Match device tags",
        related_name="+",
        blank=True,
    )
    match_fhrpgroup_tags = models.ManyToManyField(
        to=Tag,
        help_text="Use this zone for FHRP groups with these tags",
        verbose_name="Match FHRP tags",
        related_name="+",
        blank=True,
    )
    match_device_roles = models.ManyToManyField(
        to=DeviceRole,
        help_text="Use this zone for devices or VMs with these roles",
        verbose_name="Match device roles",
        related_name="+",
        blank=True,
    )
    match_interface_mgmt_only = models.BooleanField(
        default=False,
        verbose_name="Match management interfaces only",
        help_text="Use this zone only for addresses assigned to management interfaces (other criteria must match first)"
    )
    naming_ip_method = models.CharField(
        max_length=200,
        help_text="How to construct DNS name from IP address",
        verbose_name="IP naming method",
        choices=NamingIpChoices,
        default=None,
        null=True,
    )
    naming_device_method = models.CharField(
        max_length=200,
        help_text="How to construct DNS name from device IP is assigned to",
        verbose_name="Device naming method",
        choices=NamingDeviceChoices,
        default=None,
        null=True,
    )
    naming_fgrpgroup_method = models.CharField(
        max_length=200,
        help_text="How to construct DNS name from FHRP Group IP is assigned to",
        verbose_name="FHRP Group naming method",
        choices=NamingFgrpGroupChoices,
        default=None,
        null=True,
    )
    # netbox-plugin-dns also has Zone model
    # if both plugins are installed, django complains:
    #   netbox_dns.Zone.tags: (fields.E304) Reverse accessor 'Tag.zone_set'
    #   for 'netbox_dns.Zone.tags' clashes with reverse accessor for
    #   'netbox_powerdns_sync.Zone.tags'
    # So let's disable generating reverse relation here.
    tags = TaggableManager(
        through="extras.TaggedItem",
        related_name="+"
    )

    @property
    def is_reverse(self) -> bool:
        if not self.name:
            return False
        return is_reverse(self.name)

    objects = ZoneQuerySet.as_manager()

    def clean_fields(self, exclude=None):
        if self.is_reverse and self.is_default:
            msg = "Reverse zone cannot be set as default"
            if exclude and "is_default" in exclude:
                raise ValidationError(msg)
            else:
                raise ValidationError({"is_default": msg})
        
        if not self.is_reverse and not any((
            self.naming_ip_method, self.naming_fgrpgroup_method, self.naming_device_method
        )):
            raise ValidationError("At least one of naming methods must be set")

    def delete(self, *args, **kwargs):
        # delete any scheduled jobs for this zone
        if self.pk:
            jobs = Job.objects.filter(
                object_type_id=ContentType.objects.get_for_model(self).pk,
                object_id=self.pk,
                status="scheduled",
                name=JOB_NAME_SYNC
            )
            jobs.delete()
        return super().delete(*args, **kwargs)

    @classmethod
    def get_best_zone(cls, name: str) -> "Zone|None":
        """ Get the longest matching zone for a given name """
        name = make_canonical(name)
        best_match = None
        for zone in cls.objects.all():
            if name.endswith(zone.name):
                if not best_match or len(best_match.name) < len(zone.name):
                    best_match = zone
        return best_match

    @classmethod
    def match_ip(cls, ip:IPAddress) -> "[Zone]":
        """
        For a given IPAddress figure out the correct zones based on
        IPAddress, Interface, Device ir FHRPGroup tags or Device role.
        """
        def filter_mgmt_only(zones, ip):
            """ Checks if IP is assigned to interface.mgmt_only=True and finds matching zones """
            if not zones:
                # no results, just return
                return zones
            if not any(zones.values_list("match_interface_mgmt_only", flat=True)):
                # no zones are constrained to mgmt_only
                return zones
            if ip.interface.filter(mgmt_only=True):
                # IP is assigned to mgmt_only interface, return matching zones
                return zones.filter(match_interface_mgmt_only=True)
            else:
                # or just return zones
                return zones

        # tag on IPAddress
        zones = cls.objects.filter(match_ipaddress_tags__in=ip.tags.all())
        zones = filter_mgmt_only(zones, ip)
        if zones:
            return zones
        if isinstance(ip.assigned_object, Interface) or \
            isinstance(ip.assigned_object, VMInterface):
            # tag on Interface or VMInterface
            zones = cls.objects.filter(match_interface_tags__in=ip.assigned_object.tags.all())
            zones = filter_mgmt_only(zones, ip)
            if zones:
                return zones
            # tag on Device or VirtualMachine
            host = get_ip_host(ip)
            zones = cls.objects.filter(match_device_tags__in=host.tags.all())
            zones = filter_mgmt_only(zones, ip)
            if zones:
                return zones
        if isinstance(ip.assigned_object, FHRPGroup):
            zones = cls.objects.filter(match_fhrpgroup_tags__in=ip.assigned_object.tags.all())
            zones = filter_mgmt_only(zones, ip)
            if zones:
                return zones
        host = get_ip_host(ip)
        role = None
        if host:
            role = host.role
        if role:
            zones = cls.objects.filter(match_device_roles__in=[role])
            zones = filter_mgmt_only(zones, ip)
            if zones:
                return zones
        # default zone
        zones = cls.objects.filter(is_default=True)
        zones = filter_mgmt_only(zones, ip)
        if zones:
            return zones
        return cls.objects.none()

    clone_fields = (
        "name", "description", "api_servers", "default_ttl",
        "match_ipaddress_tags", "match_interface_tags", "match_device_tags",
        "match_fhrpgroup_tags", "match_device_roles", "match_interface_mgmt_only",
        "naming_ip_method", "naming_device_method", "naming_fgrpgroup_method",
        "tags",
    )

    class Meta:
        ordering = ("name",)
        constraints = [
            models.UniqueConstraint(
                fields=["is_default"],
                condition=models.Q(is_default=True),
                name="unique_is_default",
                violation_error_message="Only one zone can be set as default",
            )
        ]

    def get_absolute_url(self):
        return reverse("plugins:netbox_powerdns_sync:zone", args=[self.pk])

    def __str__(self):
        return self.name



import logging

from django.dispatch import receiver
from django.db import transaction
from django.db.models.signals import post_save

from core.models import Job
from dcim.models import Device, Interface
from extras.plugins.utils import get_plugin_config
from ipam.models import IPAddress, FHRPGroup
from netbox.context import current_request
from virtualization.models import VirtualMachine, VMInterface

from .constants import JOB_NAME_DEVICE, JOB_NAME_INTERFACE, JOB_NAME_IP, PLUGIN_NAME
from .jobs import PowerdnsTaskIP
from .utils import find_objectchange_ip


logger = logging.getLogger("netbox.netbox_powerdns_sync.signals")

IPADDRESS_DNS_FIELDS = (
    "address",
    "dns_name",
    "assigned_object_id",
    "assigned_object_type",
)

INTERFACE_DNS_FIELDS = (
    "name",
)

DEVICE_DNS_FIELDS = (
    "name",
    "primary_ip4",
    "primary_ip6",
)


@receiver(post_save, sender=IPAddress)
def update_ipaddress_dns(instance, **kwargs):
    if not get_plugin_config(PLUGIN_NAME, "post_save_enabled"):
        return
    changed = True
    if hasattr(instance, "_prechange_snapshot"):
        prechange_snapshot = instance._prechange_snapshot
        postchange_snapshot = instance.serialize_object()
        # determine if important fields changed
        changed = False
        for field in IPADDRESS_DNS_FIELDS:
            if prechange_snapshot.get(field) != postchange_snapshot.get(field):
                changed = True
    if not changed:
        # nothing interesting changed, nothing to do
        return
    job_args = dict(
        func=PowerdnsTaskIP.run_update_ip,
        instance=instance,
        name=JOB_NAME_IP,
    )
    request = current_request.get()
    if request:
        job_args["user"] = request.user
    transaction.on_commit(lambda: Job.enqueue(**job_args))


@receiver(post_save, sender=Interface)
@receiver(post_save, sender=VMInterface)
@receiver(post_save, sender=FHRPGroup)
def update_interface_dns(instance: FHRPGroup, **kwargs):
    if not get_plugin_config(PLUGIN_NAME, "post_save_enabled"):
        return
    changed = True
    if hasattr(instance, "_prechange_snapshot"):
        prechange_snapshot = instance._prechange_snapshot
        postchange_snapshot = instance.serialize_object()
        # determine if relevant fields changed
        changed = False
        for field in INTERFACE_DNS_FIELDS:
            if prechange_snapshot.get(field) != postchange_snapshot.get(field):
                changed = True
    if not changed:
        # nothing interesting changed, nothing to do
        return
    # need to create job for IPv4 and IPv6
    job_args = dict(
        func=PowerdnsTaskIP.run_update_ip,
        name=JOB_NAME_INTERFACE,
    )
    request = current_request.get()
    if request:
        job_args["user"] = request.user
    for ip_address in instance.ip_addresses.all():
        job_args["instance"] = ip_address
        transaction.on_commit(lambda: Job.enqueue(**job_args))


@receiver(post_save, sender=Device)
@receiver(post_save, sender=VirtualMachine)
def update_device_dns(instance, **kwargs):
    if not get_plugin_config(PLUGIN_NAME, "post_save_enabled"):
        return
    changed = False
    if hasattr(instance, "_prechange_snapshot"):
        prechange_snapshot = instance._prechange_snapshot
        postchange_snapshot = instance.serialize_object()
        # determine if relevant fields changed
        changed = False
        for field in DEVICE_DNS_FIELDS:
            if prechange_snapshot.get(field) != postchange_snapshot.get(field):
                changed = True
    if not changed:
        # nothing interesting changed, nothing to do
        return
    request = current_request.get()
    # need to create job for IPv4 and IPv6
    if instance.primary_ip4:
        enqueue_task = True
        job_args = dict(
            func=PowerdnsTaskIP.run_update_ip,
            instance=instance.primary_ip4,
            name=JOB_NAME_DEVICE,
        )
        if request:
            job_args["user"] = request.user
            # check if IPAddress object was creted in same request
            q_created = find_objectchange_ip(instance.primary_ip4, request.id)
            enqueue_task = not q_created.exists()
        if enqueue_task:
            transaction.on_commit(lambda: Job.enqueue(**job_args))
    if instance.primary_ip6:
        enqueue_task = True
        job_args = dict(
            func=PowerdnsTaskIP.run_update_ip,
            instance=instance.primary_ip6,
            name=JOB_NAME_DEVICE,
        )
        if request:
            job_args["user"] = request.user
            # check if IPAddress object was creted in same request
            q_created = find_objectchange_ip(instance.primary_ip6, request.id)
            enqueue_task = not q_created.exists()
        if enqueue_task:
            transaction.on_commit(lambda: Job.enqueue(**job_args))



from django.db.models import Q
from utilities.querysets import RestrictedQuerySet

from .constants import PTR_ZONE_SUFFIXES


class EnabledQuerySet(RestrictedQuerySet):
    def enabled(self):
        return self.filter(enabled=True)


class ZoneQuerySet(EnabledQuerySet):
    def _gen_q(self):
        q = Q()
        for s in PTR_ZONE_SUFFIXES:
            q |= Q(name__endswith=s)
        return q

    def forward(self):
        return self.exclude(self._gen_q())

    def reverse(self):
        return self.filter(self._gen_q())



from netbox.views import generic
from utilities.utils import count_related
from utilities.views import register_model_view

from .. import filtersets, forms, tables
from ..models import ApiServer, Zone

__all__ = (
    "ApiServerListView",
    "ApiServerView",
    "ApiServerEditView",
    "ApiServerDeleteView",
    #"ApiServerBulkImportView",
    #"ApiServerBulkEditView",
    "ApiServerBulkDeleteView",
)

class ApiServerListView(generic.ObjectListView):
    queryset = ApiServer.objects.annotate(
        zone_count=count_related(Zone, 'api_servers'),
    )
    filterset = filtersets.ApiServerFilterSet
    filterset_form = forms.ApiServerFilterForm
    table = tables.ApiServerTable


@register_model_view(ApiServer)
class ApiServerView(generic.ObjectView):
    queryset = ApiServer.objects.all()

    def get_extra_context(self, request, instance):
        zones = Zone.objects.restrict(request.user, 'view').filter(api_servers=instance)
        zone_table = tables.ZoneTable(zones, user=request.user)
        zone_table.columns.hide('api_server_count')
        zone_table.configure(request)

        return {
            'zones': zones,
            'zone_table': zone_table,
        }


@register_model_view(ApiServer, 'edit')
class ApiServerEditView(generic.ObjectEditView):
    queryset = ApiServer.objects.all()
    form = forms.ApiServerForm


@register_model_view(ApiServer, 'delete')
class ApiServerDeleteView(generic.ObjectDeleteView):
    queryset = ApiServer.objects.all()


# class ApiServerBulkImportView(generic.BulkImportView):
    # queryset = ApiServer.objects.all()
    # model_form = forms.ApiServerImportForm


# class ApiServerBulkEditView(generic.BulkEditView):
    # queryset = ApiServer.objects.annotate(
        # zone_count=count_related(Zone, 'api_servers'),
    # )
    # filterset = filtersets.ApiServerFilterSet
    # table = tables.ApiServerTable
    # form = forms.ApiServerBulkEditForm


class ApiServerBulkDeleteView(generic.BulkDeleteView):
    queryset = ApiServer.objects.annotate(
        zone_count=count_related(Zone, 'api_servers'),
    )
    filterset = filtersets.ApiServerFilterSet
    table = tables.ApiServerTable



import importlib
from dcim.models import Interface
from ipam.models import IPAddress, FHRPGroup
from virtualization.models import VMInterface

from .models import Zone
from .utils import make_dns_label, make_canonical


def _load_class(path:str) -> object:
    try:
        module_path, class_name = path.rsplit(".", maxsplit=1)
        module = importlib.import_module(module_path)
        klass = getattr(module, class_name)
        return klass
    except:
        return None


def generate_fqdn(ip: IPAddress, zone:Zone) -> str|None:
    fqdn = None
    if not zone:
        return None
    if zone.naming_ip_method:
        klass = _load_class(zone.naming_ip_method)
        naming_method = klass(ip, zone)
        fqdn = naming_method.make_fqdn()
    if not fqdn and zone.naming_device_method:
        klass = _load_class(zone.naming_device_method)
        naming_method = klass(ip, zone)
        fqdn = naming_method.make_fqdn()
    if not fqdn and zone.naming_fgrpgroup_method:
        klass = _load_class(zone.naming_fgrpgroup_method)
        naming_method = klass(ip, zone)
        fqdn = naming_method.make_fqdn()
    return fqdn


class NamingBase:
    def __init__(self, ip:IPAddress, zone:Zone):
        self.ip = ip
        self.zone = zone
        self.interface = None
        self.host = None
    
    def make_fqdn(self) -> str|None:
        fqdn = None
        name = self.make_name()
        if not name:
            return None
        name = make_canonical(name)
        # check if name already includes zone name
        if name.endswith(self.zone.name):
            fqdn = name
        else:
            fqdn = name + self.zone.name
        return fqdn

    def make_name(self) -> str|None:
        raise NotImplementedError()

    def _populate_host_interface(self):
        if isinstance(self.ip.assigned_object, Interface):
            self.interface = self.ip.assigned_object
            self.host = self.ip.assigned_object.device
        elif isinstance(self.ip.assigned_object, VMInterface):
            self.interface = self.ip.assigned_object
            self.host = self.ip.assigned_object.virtual_machine


class NamingDeviceByInterfacePrimary(NamingBase):
    """
    Generate name in formatted as: interface.device.zone
    If IP is primary, don't prepend interface name
    """
    def make_name(self) -> str|None:
        self._populate_host_interface()
        if self.host:
            name = self.host.name
            name = ".".join(map(make_dns_label, name.split(".")))
            if self.ip != self.host.primary_ip4 and self.ip != self.host.primary_ip6:
                name = make_dns_label(self.interface.name) + "." + name
            return name


class NamingDeviceByInterface(NamingBase):
    """ Generate name in formatted as: interface.device.zone """
    def make_name(self) -> str|None:
        self._populate_host_interface()
        if self.host:
            name = self.host.name
            name = ".".join(map(make_dns_label, name.split(".")))
            name = make_dns_label(self.interface.name) + "." + name
            return name


class NamingDeviceName(NamingBase):
    """ Generate name as: device.zone """
    def make_name(self) -> str|None:
        self._populate_host_interface()
        if self.host:
            name = self.host.name
            name = ".".join(map(make_dns_label, name.split(".")))
            return name


class NamingIpDnsName(NamingBase):
    """ Use IPAddress.dns_name """
    def make_name(self) -> str|None:
        if self.ip.dns_name:
            return self.ip.dns_name


class NamingIpReverse(NamingBase):
    """ Use IPAddress in reverse foramt """
    def make_name(self) -> str|None:
        return ".".join(self.ip.address.ip.reverse_dns.split(".")[:-3])


class NamingFGRPGroupName(NamingBase):
    """ Use FGRPGroup name: fgrp-group.zone """
    def make_name(self) -> str|None:
        if isinstance(self.ip.assigned_object, FHRPGroup):
            name = self.ip.assigned_object.name
            name = ".".join(map(make_dns_label, name.split(".")))



from rest_framework import serializers

from netbox.api.serializers import WritableNestedSerializer
from ..models import ApiServer, Zone

__all__ = (
    "NestedApiServerSerializer",
    "NestedZoneSerializer",
)


class NestedApiServerSerializer(WritableNestedSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netbox_powerdns_sync-api:apiserver-detail"
    )

    class Meta:
        model = ApiServer
        fields = ("id", "url", "display", "name", "api_url", "api_token", "enabled")


class NestedZoneSerializer(WritableNestedSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netbox_powerdns_sync-api:zone-detail"
    )

    class Meta:
        model = Zone
        fields = ("id", "url", "display", "name", "enabled")



from rest_framework import serializers
from netbox.api.serializers import NetBoxModelSerializer, NestedTagSerializer
from dcim.api.serializers import NestedDeviceRoleSerializer

from .nested_serializers import *
from ..models import ApiServer, Zone


class ApiServerSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netbox_powerdns_sync-api:apiserver-detail"
    )
    zones = NestedZoneSerializer(many=True, required=False, read_only=True)

    class Meta:
        model = ApiServer
        fields = (
            "id", "url", "display", "name", "api_url", "api_token", "enabled",
            "description", "zones", "tags", "custom_fields", "created", "last_updated"
        )


class ZoneSerializer(NetBoxModelSerializer):
    url = serializers.HyperlinkedIdentityField(
        view_name="plugins-api:netbox_powerdns_sync-api:zone-detail"
    )
    is_reverse = serializers.BooleanField(read_only=True)
    api_servers = NestedApiServerSerializer(many=True, required=True)
    match_ipaddress_tags = NestedTagSerializer(many=True, required=False)
    match_interface_tags = NestedTagSerializer(many=True, required=False)
    match_device_tags = NestedTagSerializer(many=True, required=False)
    match_fhrpgroup_tags = NestedTagSerializer(many=True, required=False)
    match_device_roles = NestedDeviceRoleSerializer(many=True, required=False)

    class Meta:
        model = Zone
        fields = (
            "id", "url", "display", "name", "description", "api_servers", "enabled",
            "is_reverse", "is_default", "default_ttl", "match_ipaddress_tags",
            "match_interface_tags", "match_device_tags", "match_fhrpgroup_tags",
            "match_device_roles", "match_interface_mgmt_only", "naming_ip_method",
            "naming_device_method", "naming_fgrpgroup_method", "tags",
            "custom_fields", "created", "last_updated"
        )



from django import forms
from dcim.models import DeviceRole
from netbox.forms import NetBoxModelFilterSetForm
from utilities.forms import BOOLEAN_WITH_BLANK_CHOICES
from utilities.forms.fields import DynamicModelMultipleChoiceField, TagFilterField

from ..choices import NamingDeviceChoices, NamingFgrpGroupChoices, NamingIpChoices
from ..fields import MatchTagFilterField
from ..models import ApiServer, Zone


__all__ = (
    "ApiServerFilterForm",
    "ZoneFilterForm",
)


class ApiServerFilterForm(NetBoxModelFilterSetForm):
    model = ApiServer

    zone_id = DynamicModelMultipleChoiceField(
        queryset=Zone.objects.all(),
        required=False,
        null_option="None",
        label="Zone",
    )
    enabled = forms.NullBooleanField(
        required=False,
        label="Is Enabled",
        widget=forms.Select(
            choices=BOOLEAN_WITH_BLANK_CHOICES
        )
    )
    tag = TagFilterField(model)


class ZoneFilterForm(NetBoxModelFilterSetForm):
    model = Zone

    api_server_id = DynamicModelMultipleChoiceField(
        queryset=ApiServer.objects.all(),
        required=False,
        null_option="None",
        label="API Server",
    )
    enabled = forms.NullBooleanField(
        required=False,
        label="Is Enabled",
        widget=forms.Select(
            choices=BOOLEAN_WITH_BLANK_CHOICES
        )
    )
    reverse = forms.NullBooleanField(
        required=False,
        label="Reverse zones",
        widget=forms.Select(
            choices=BOOLEAN_WITH_BLANK_CHOICES
        )
    )
    match_tags = MatchTagFilterField()
    match_device_roles = DynamicModelMultipleChoiceField(
        queryset=DeviceRole.objects.all(),
        required=False,
        null_option="None",
        label="Match Device Roles",
    )
    naming_ip_method = forms.MultipleChoiceField(
        required=False,
        label="IP naming method",
        choices=NamingIpChoices,
    )
    naming_device_method = forms.MultipleChoiceField(
        required=False,
        label="Device naming method",
        choices=NamingDeviceChoices,
    )
    naming_fgrpgroup_method = forms.MultipleChoiceField(
        required=False,
        label="FHRP Group naming method",
        choices=NamingFgrpGroupChoices,
    )
    tag = TagFilterField(model)



from netbox.api.routers import NetBoxRouter
from . import views


app_name = 'netbox_powerdns_sync'

router = NetBoxRouter()
router.register('api-servers', views.ApiServerViewSet)
router.register('zones', views.ZoneViewSet)

urlpatterns = router.urls



from django import forms
from extras.forms import ScriptForm
from utilities.forms.fields import DynamicModelMultipleChoiceField

from ..models import Zone


class ZoneScheduleForm(ScriptForm):
    zones = DynamicModelMultipleChoiceField(
        queryset=Zone.objects.enabled(),
        query_params={"enabled": True},
        help_text="Only enabled zones can be scheduled",
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields.pop("_commit")
        self.fields["_schedule_at"].help_text = self.fields["_schedule_at"].help_text.replace("script", "sync")
        self.fields["_interval"].help_text = self.fields["_interval"].help_text.replace("script", "sync")



import re
import unicodedata
from powerdns import Comment, RRSet
from django.contrib.contenttypes.models import ContentType
from dcim.models import Device, Interface
from extras.choices import ObjectChangeActionChoices
from extras.plugins.utils import get_plugin_config
from extras.models import ObjectChange
from ipam.models import IPAddress
from virtualization.models import VirtualMachine, VMInterface

from .constants import PLUGIN_NAME, FAMILY_TYPES, PTR_TYPE, PTR_ZONE_SUFFIXES


def get_ip_host(ip:IPAddress) -> Device|VirtualMachine|None:
    """
    If IPAddress is assigned to Interface or VMInterface return Device
    or VirtualMachine
    """
    if isinstance(ip.assigned_object, Interface):
        return ip.assigned_object.device
    if isinstance(ip.assigned_object, VMInterface):
        return ip.assigned_object.virtual_machine
    return None


def get_ip_ttl(ip: IPAddress) -> int|None:
    """
    Get TTL from IPAddress custom field if set. Else None.
    """
    ttl = None
    ttl_cf = get_plugin_config(PLUGIN_NAME, "ttl_custom_field")
    if ttl_cf and ttl_cf in ip.cf:
        ttl = ip.cf.get(ttl_cf)
        if ttl and not isinstance(ttl, int):
            raise ValueError(f"Custom field {ttl_cf} must be a positive integer")
        if ttl and ttl < 1:
            raise ValueError(f"Custom field {ttl_cf} must be a positive integer")
    return ttl


def can_manage_record(record: dict|RRSet) -> bool:
    """
    Check if record from powerdns is of supported type and if using comments,
    check if it's correct
    """
    comment = get_plugin_config(PLUGIN_NAME, "powerdns_managed_record_comment")
    managed_types = [PTR_TYPE] + list(FAMILY_TYPES.values())
    if record["type"] not in managed_types:
        return False
    if comment:
        for record_comment in record["comments"]:
            if record_comment["content"] == comment:
                return True
        return False
    return True


def get_managed_comment() -> list:
    """
    Return powerdns RRset comment if set in confiuration
    """
    comment = get_plugin_config(PLUGIN_NAME, "powerdns_managed_record_comment")
    if comment:
        return [Comment(comment)]
    else:
        return []


def make_canonical(name: str) -> str:
    """ Ensure string ends with a dot """
    if name[-1] != ".":
        name = name + "."
    return name


def make_dns_label(name: str) -> str:
    """
    Convert to ASCII. Convert spaces, dosts or slashes or repeated dashes to
    single dashes. Remove characters that aren't alphanumerics, underscores,
    hyphens or slashes. Convert to lowercase. Also strip leading and trailing
    whitespace, dashes, and underscores.
    (adapted from django's slugify function)
    """
    name = unicodedata.normalize("NFKD", name).encode("ascii", "ignore").decode("ascii")
    name = re.sub(r"[^\w\s\-\/\._]", "", name.lower())
    return re.sub(r"[_\.\/\-\s]+", "-", name).strip("-_")


def is_reverse(name:str) -> bool:
    return any(map(lambda s: name.endswith(s), PTR_ZONE_SUFFIXES))


def find_objectchange_ip(ip, request_id):
    return ObjectChange.objects.filter(
        action=ObjectChangeActionChoices.ACTION_CREATE,
        request_id=request_id,
        changed_object_type=ContentType.objects.get_for_model(ip),
        changed_object_id=ip.pk,
    )



from netbox.views import generic
from utilities.utils import count_related
from utilities.views import register_model_view

from .. import filtersets, forms, tables
from ..models import ApiServer, Zone

__all__ = (
    "ZoneListView",
    "ZoneView",
    "ZoneEditView",
    "ZoneDeleteView",
    #"ZoneBulkImportView",
    #"ZoneBulkEditView",
    "ZoneBulkDeleteView",
)

class ZoneListView(generic.ObjectListView):
    queryset = Zone.objects.annotate(
        api_server_count=count_related(ApiServer, 'zones'),
    )
    filterset = filtersets.ZoneFilterSet
    filterset_form = forms.ZoneFilterForm
    table = tables.ZoneTable


@register_model_view(Zone)
class ZoneView(generic.ObjectView):
    queryset = Zone.objects.all()


@register_model_view(Zone, 'edit')
class ZoneEditView(generic.ObjectEditView):
    queryset = Zone.objects.all()
    form = forms.ZoneForm


@register_model_view(Zone, 'delete')
class ZoneDeleteView(generic.ObjectDeleteView):
    queryset = Zone.objects.all()


# class ZoneBulkImportView(generic.BulkImportView):
    # queryset = Zone.objects.all()
    # model_form = forms.ZoneImportForm


# class ZoneBulkEditView(generic.BulkEditView):
    # queryset = Zone.objects.annotate(
        # zone_count=count_related(Zone, 'api_servers'),
    # )
    # filterset = filtersets.ZoneFilterSet
    # table = tables.ZoneTable
    # form = forms.ZoneBulkEditForm


class ZoneBulkDeleteView(generic.BulkDeleteView):
    queryset = Zone.objects.all()
    filterset = filtersets.ZoneFilterSet
    table = tables.ZoneTable



import django_filters
from django.db.models import Q
from dcim.models import DeviceRole
from netbox.filtersets import NetBoxModelFilterSet
from utilities import filters

from .choices import NamingDeviceChoices, NamingIpChoices, NamingFgrpGroupChoices
from .models import *

__all__ = (
    "ApiServerFilterSet",
    "ZoneFilterSet",
)


class ApiServerFilterSet(NetBoxModelFilterSet):
    zone_id = django_filters.ModelMultipleChoiceFilter(
        queryset=Zone.objects.all(),
        field_name="zones",
        label="Zone (ID)",
    )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value) |
            Q(api_url__icontains=value)
        )

    class Meta:
        model = ApiServer
        fields = ["id", "name", "description", "api_url", "enabled"]


class ZoneFilterSet(NetBoxModelFilterSet):
    api_server_id = django_filters.ModelMultipleChoiceFilter(
        queryset=ApiServer.objects.all(),
        field_name="api_servers",
        label="API Server (ID)",
    )
    reverse = django_filters.BooleanFilter(
        method="filter_reverse",
        label="Reverse zones"
    )
    match_tags = filters.MultiValueCharFilter(
        method="filter_match_tags",
        label="Match Tags",
    )
    match_device_roles = django_filters.ModelMultipleChoiceFilter(
        queryset=DeviceRole.objects.all(),
        label="Match Device Roles (ID)",
    )
    naming_ip_method = django_filters.MultipleChoiceFilter(
        choices=NamingIpChoices,
    )
    naming_device_method = django_filters.MultipleChoiceFilter(
        choices=NamingDeviceChoices,
    )
    naming_fgrpgroup_method = django_filters.MultipleChoiceFilter(
        choices=NamingFgrpGroupChoices,
    )

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset
        return queryset.filter(
            Q(name__icontains=value)|
            Q(description__icontains=value)
        )

    def filter_match_tags(self, queryset, name, value):        
        query = Q(match_ipaddress_tags__slug__in=value)|\
            Q(match_interface_tags__slug__in=value)|\
            Q(match_device_tags__slug__in=value)|\
            Q(match_fhrpgroup_tags__slug__in=value)
        return queryset.filter(query).distinct()

    def filter_reverse(self, queryset, name, value):
        if value:
            return queryset.reverse()
        else:
            return queryset.forward()

    class Meta:
        model = Zone
        fields = ["id", "name", "description", "enabled"]



from django.urls import include, path

from utilities.urls import get_model_urls
from . import models, views


urlpatterns = (

    # API Servers
    path('api-servers/', views.ApiServerListView.as_view(), name='apiserver_list'),
    path('api-servers/add/', views.ApiServerEditView.as_view(), name='apiserver_add'),
    #path('api-servers/import/', views.ApiServerBulkImportView.as_view(), name='apiserver_import'),
    #path('api-servers/edit/', views.ApiServerBulkEditView.as_view(), name='apiserver_bulk_edit'),
    path('api-servers/delete/', views.ApiServerBulkDeleteView.as_view(), name='apiserver_bulk_delete'),
    path('api-servers/<int:pk>/', include(get_model_urls('netbox_powerdns_sync', 'apiserver'))),

    # Zones
    path('zones/', views.ZoneListView.as_view(), name='zone_list'),
    path('zones/add/', views.ZoneEditView.as_view(), name='zone_add'),
    #path('zones/import/', views.ZoneBulkImportView.as_view(), name='zone_import'),
    #path('zones/edit/', views.ZoneBulkEditView.as_view(), name='zone_bulk_edit'),
    path('zones/delete/', views.ZoneBulkDeleteView.as_view(), name='zone_bulk_delete'),
    path('zones/<int:pk>/', include(get_model_urls('netbox_powerdns_sync', 'zone'))),

    path('sync/', views.SyncJobsView.as_view(), name='sync_jobs'),
    path('sync/schedule/', views.SyncScheduleView.as_view(), name='sync_schedule'),
    path('sync/<int:job_pk>/', views.SyncResultView.as_view(), name='sync_result'),

)



from extras.models import Tag


from django import forms
from django.db.models import Count


class MatchTagFilterField(forms.MultipleChoiceField):
    """
    A filter field for tags. Filtering is done on tag.slug
    """

    def __init__(self, *args, **kwargs):
        def get_choices():
            tags = Tag.objects.annotate(
                count=Count('extras_taggeditem_items')
            ).order_by('name')
            return [
                (str(tag.slug), '{} ({})'.format(tag.name, tag.count)) for tag in tags
            ]

        # Choices are fetched each time the form is initialized
        super().__init__(label='Match Tags', choices=get_choices, required=False, *args, **kwargs)



from netbox.api.viewsets import NetBoxModelViewSet
from .. import filtersets, models
from .serializers import ApiServerSerializer, ZoneSerializer


class ApiServerViewSet(NetBoxModelViewSet):
    queryset = models.ApiServer.objects.prefetch_related("tags", "zones")
    serializer_class = ApiServerSerializer
    filterset_class = filtersets.ApiServerFilterSet


class ZoneViewSet(NetBoxModelViewSet):
    queryset = models.Zone.objects.prefetch_related(
        "api_servers",
        "match_ipaddress_tags",
        "match_interface_tags",
        "match_device_tags",
        "match_fhrpgroup_tags",
        "match_device_roles",
        "tags"
    )
    serializer_class = ZoneSerializer
    filterset_class = filtersets.ZoneFilterSet



class PowerdnsSyncNoNameFound(Exception):
    pass


class PowerdnsSyncNoZoneFound(Exception):
    pass


class PowerdnsSyncServerError(Exception):
    pass


class PowerdnsSyncServerZoneMissing(PowerdnsSyncServerError):
    pass


class PowerdnsSyncNoServers(PowerdnsSyncServerError):
    pass
